	REPEAT		=  2
	VIRTUAL		= 0x80000000
	FP		= 62
	SP		= 63
	FLAGS		= 74
	PC		= 77
	CODE		= 78
	DATA		= 79
	T_IDIVZ		=  0
	T_IOVFL		=  1
	T_SUPER		= 10
	T_MEMFLT	= 16
	T_PAGEFLT	= 17
	T_INVAL		= 18
	T_U_EXIT	= 24
	PAGE_BITS	= 12
	PAGE_SIZE	= (1 << PAGE_BITS)

start:
	cmpe	%1, %1, -1843		! Memory dump magic
init:
	mov	%0, 01000		! I'm written to version 1.0
	mov	%1, 5			! Number of ASIDs please
	trap	%52, %0, -4		! #ASIDs in %52
	mov	%50, REPEAT		! How often to run this test?
	mov	%51, 1			! ASID to use for test program
run:
	mov	%0, %code
	trap	%10, %0, -2		! Where the heck am I?
	sub	%20, %14, %12		! Physical minus Virtual (V2P) for me
	lea	%22, [c_prog1]		! Virtual starting address of program
	mov	%0, PAGE_SIZE
	add	%21, %20, %0		! My virtual PAGE_SIZE to physical
	sub	%21, %22		! V2P offset for the program

	sub	%0, %sp, 4*128*W	! Other program context below my stack
	mov	%1, ~(PAGE_SIZE-1)
	and	%0, %1			! Align down to the page size
	mov	%1, %0
	mov	%2, 0
	mov	%3, 128*W
	fill	%1, %2, %3		! Clear context

	add	%10, %0, %20		! %0 to physical
	sub	%10, %21		! %10 = %0 + myV2P - progV2P
	st	%10, [%0 + SP*W]	! %sp = starts below context
	st	%22, [%0 + PC*W]	! %pc = prog1
	st	%51, [%0 + CODE*W]	! Code address space ID
	st	%51, [%0 + DATA*W]	! Common I&D

	mov	%10, %51		! Address space of program
	mov	%11, 0			! Not present, so no linear area
	mov	%12, 0			! Base virtual address 0
	mov	%13, 0			! Size zero
	mov	%14, 0			! Physical address 0
	trap	%10, %10, -1		! Create new address space

	mov	%10, PAGE_SIZE
	sub	%23, %0, %10		! Program's available memory size

trap_handler:
	queue	%0, 0			! Put program on the run queue
	queue	%0, 3			! Wait for an exception
	cmpe	%10, %1, T_PAGEFLT	! Page faults are most likely
	jf	%10, odd_trap
	sub	%10, %3, %22		! %10 = fault address - prog1
	cmpb	%10, %10, %23		! %10 = (faddr - prog1) < prog1size
	jf	%10, bad_page
	mov	%10, %2			! Address space
	mov	%11, 0xF		! PRWX
	mov	%12, %3			! Virtual = Offending address
	add	%13, %12, %21		! Physical = Virtual + V2P
	trap	%10, %10, -3		! Load TLB entry
	jmp	trap_handler		! Continue where we left off
odd_trap:
	cmpe	%10, %1, T_U_EXIT	! Done?
	jt	%10, exit
	cmpbe	%10, %1, T_SUPER	! Ordinary trap?
	jf	%10, bad_trap
	cmpe	%10, %1, T_SUPER	! Supervisor reserved trap?
	jf	%10, 0f
	ld	%10, [%0 + FLAGS*W]
	bic	%10, T_SUPER		! Be nice, reenable supervisor mode
	st	%10, [%0 + FLAGS*W]
0:	st	%1, [%0 + 61*W]		! Program's %61 = trap number
	jmp	trap_handler
bad_trap:
	trap	%1, %1, T_INVAL		! Trap we can't handle
bad_page:
	trap	%1, %1, T_PAGEFLT	! Outside program error
exit:
	add	%51, 1			! Use another ASID
	cmpb	%0, %51, %52		! Below #ASIDs?
	jt	%0, 0f
	mov	%51, 1			! Address space 1
0:	sub	%50, 1			! Once more?
	jnz	%50, run
	mov	%0, 0
	trap	%0, %0, -7		! We're done

	align	W
c_prog1:dataw	prog1-.

	align	PAGE_SIZE		! Put in a fresh page
	org	VIRTUAL			! Virtual address of program
prog1:	! Plug in any simple program here:
!******************************************************************************
test:
	jmp	0f			! 00
bad:	ill

trap_function:				! User trap function.
	ld	%60, [%sp+1*W]		! Put the trap number in %60
	ret	2*W			! Return and pop type+op
0:
	mov	%10, -1
	mov	%11, 0
	mov	%12, 1
	jl	%10, 0f			! 01
	ill
0:	jl	%11, bad
	jl	%12, bad

	jg	%10, bad		! 02
	jg	%11, bad
	jg	%12, 0f
	ill
0:
	jle	%10, 0f			! 03
	ill
0:	jle	%11, 0f
	ill
0:	jle	%12, bad

	jge	%10, bad		! 04
	jge	%11, 0f
	ill
0:	jge	%12, 0f
	ill
0:
	je	%10, bad		! 05
	je	%11, 0f
	ill
0:	je	%12, bad

	jne	%10, 0f			! 06
	ill
0:	jne	%11, bad
	jne	%12, 0f
	ill
0:
	jmp	[0f]			! 07
	align W
0:	dataw	1f-.
	ill
1:
	lea	%0, [0f]
	jmp	%0			! 08
	align	W
0:	dataw	1f-.
	ill
1:
	mov	%0, 0
9:	jmp	8f[%0]			! 6B
8:	data4	0f-.
	data4	1f-.
	data4	2f-.
0:	cmpe	%1, %0, 0
	jf	%1, bad
	jmp	7f
1:	cmpe	%1, %0, 1
	jf	%1, bad
	jmp	7f
2:	cmpe	%1, %0, 2
	jf	%1, bad
	jmp	6f
7:	add	%0, 1
	jmp	8b[%0]
6:

	call	[0f]			! 09
	lea	%0, [0f]
	call	%0			! 0A
	jmp	2f
	align	W
0:	dataw	1f-.
1:	ret
2:
	mov	%0, 3
	mov	%1, 6
	mov	%10, %11, %0, %1	! 0B
	cmpe	%20, %10, %0
	jf	%20, bad
	cmpe	%20, %11, %1
	jf	%20, bad
	xchg	%0, %1
	cmpe	%20, %0, 6
	jf	%20, bad
	cmpe	%20, %1, 3
	jf	%20, bad

	!mov tested everywhere		! 0C

	mov	%fp, 123
	sub	%0, %fp, 1<<16		! 0D
	mov	%1, 123 - (1<<16)
	cmpe	%0, %0, %1
	jf	%0, bad

	add	%0, %sp, 123 - (1<<16)	! 0E
	add	%1, %sp
	cmpe	%0, %0, %1
	jf	%0, bad

	sub	%1, 123 - (1<<16)	! 0F
	cmpe	%0, %sp, %1
	jf	%0, bad

	!add tested all over		! 10
	!sub tested all over		! 11

	mov	%0, 0x32
	mov	%1, 0x11
	and	%2, %0, %1		! 12
	cmpe	%2, %2, 0x32 & 0x11
	jf	%2, bad

	andn	%2, %0, %1		! 2B
	cmpe	%2, %2, 0x32 & ~0x11
	jf	%2, bad

	or	%2, %0, %1		! 13
	cmpe	%2, %2, 0x32 | 0x11
	jf	%2, bad

	xor	%2, %0, %1		! 14
	cmpe	%2, %2, 0x32 ^ 0x11
	jf	%2, bad

	mul	%2, %0, %1		! 15
	cmpe	%2, %2, 0x32 * 0x11
	jf	%2, bad

	divu	%2, %0, %1		! 16
	cmpe	%2, %2, 0x32 / 0x11
	jf	%2, bad

	neg	%2, %0
	divs	%2, %2, %1		! 17
	cmpe	%2, %2, -0x32 / 0x11
	jf	%2, bad

	remu	%2, %0, %1		! 2C
	cmpe	%2, %2, 0x32 % 0x11
	jf	%2, bad

	neg	%2, %0
	rems	%2, %2, %1		! 2D
	cmpe	%2, %2, -0x32 % 0x11
	jf	%2, bad

	mov	%2, W*8-5
	shl	%3, %0, %2		! 18
	shr	%4, %3, %2		! 19
	cmpe	%4, %4, 0x12
	jf	%4, bad
	sar	%4, %3, %2		! 1A
	cmpe	%4, %4, -0x20+0x12
	jf	%4, bad

	mov	%2, 4
	bic	%3, %0, %2		! 2E
	cmpe	%4, %3, 0x22
	jf	%4, bad

	bis	%3, %3, %2		! 2F
	cmpe	%4, %3, 0x32
	jf	%4, bad

	bix	%3, %3, %2		! 3B
	cmpe	%4, %3, 0x22
	jf	%4, bad

	bit	%4, %0, %2		! 5C
	jf	%4, bad
	bit	%4, %4, %2
	jt	%4, bad

	!cmp* tested all over		! 21 22 25 26 28 29

	mov	%0, 0
	mov	%1, 1
	mov	%2, 2
	mov	%3, 3
	cmov	%3, %0, %2, %3		! 2A
	cmpne	%4, %3, %2
	jf	%4, bad
	cmov	%3, %1, %3, %2
	cmpne	%4, %3, %2
	jf	%4, bad
	cmov	%3, %1, %2, %3
	cmpe	%4, %3, %2
	jf	%4, bad
	cmov	%1, %0, %3, %2
	cmpe	%4, %1, %2
	jf	%4, bad

	!add, sub tested all over	! 30 31

	mov	%0, 0x32
	and	%2, %0, 0x11		! 32
	cmpe	%2, %2, 0x32 & 0x11
	jf	%2, bad

	or	%2, %0, 0x11		! 33
	cmpe	%2, %2, 0x32 | 0x11
	jf	%2, bad

	xor	%2, %0, 0x11		! 34
	cmpe	%2, %2, 0x32 ^ 0x11
	jf	%2, bad

	mul	%2, %0, 0x11		! 35
	cmpe	%2, %2, 0x32 * 0x11
	jf	%2, bad

	divu	%2, %0, 0x11		! 36
	cmpe	%2, %2, 0x32 / 0x11
	jf	%2, bad

	neg	%2, %0
	divs	%2, %2, 0x11		! 37
	cmpe	%2, %2, -0x32 / 0x11
	jf	%2, bad

	remu	%2, %0, 0x11		! 58
	cmpe	%2, %2, 0x32 % 0x11
	jf	%2, bad

	neg	%2, %0
	rems	%2, %2, 0x11		! 59
	cmpe	%2, %2, -0x32 % 0x11
	jf	%2, bad

	shl	%3, %0, W*8-5		! 38
	shr	%4, %3, W*8-5		! 39
	cmpe	%4, %4, 0x12
	jf	%4, bad
	sar	%4, %3, W*8-5		! 3A
	cmpe	%4, %4, -0x20+0x12
	jf	%4, bad

	bic	%3, %0, 4		! 5D
	cmpe	%4, %3, 0x22
	jf	%4, bad

	bis	%3, %3, 4		! 5E
	cmpe	%4, %3, 0x32
	jf	%4, bad

	bix	%3, %3, 4		! 3C
	cmpe	%4, %3, 0x22
	jf	%4, bad

	bit	%4, %0, 4		! 5F
	jf	%4, bad
	bit	%4, %4, 4
	jt	%4, bad

	!cmp* tested all over		! 41 42 45 46 48 49

	mov	%0, 0
	mov	%1, 1
	mov	%3, 3
	cmovt	%3, %0, -2		! 4A
	cmpne	%4, %3, -2
	jf	%4, bad
	cmovf	%3, %1, -2		! 4B
	cmpne	%4, %3, -2
	jf	%4, bad
	cmovt	%3, %1, -2
	cmpe	%4, %3, -2
	jf	%4, bad
	cmovf	%1, %0, -2
	cmpe	%4, %1, -2
	jf	%4, bad

	mov	%0, -10
	sub	%1, -10, %0		! 4C
	jnz	%1, bad

	mov	%0, 0
	mov	%1, 13
0:	add	%0, 1
	cjne	%0, %1, 0b		! 4D
	cmpe	%2, %0, %1
	jf	%2, bad

	ld	%0, [c0123]
	mov	%1, 0
	mov	%2, 0
0:	shr	%3, %0, %2
	and	%4, %3, 0xFF
	shl	%1, 8
	or	%1, %4
	add	%2, 8
	cmpb	%3, %2, W*8
	jt	%3, 0b
	bswap	%2, %0, 1		! 4E
	shr	%3, %1, (W-1)*8
	cmpe	%4, %2, %3
	jf	%4, bad
	bswap	%2, %0, 2
	shr	%3, %1, (W-2)*8
	cmpe	%4, %2, %3
	jf	%4, bad
	bswap	%2, %0, 4
	shr	%3, %1, (W-4)*8
	cmpe	%4, %2, %3
	jf	%4, bad
	bswap	%2, %0
	cmpe	%4, %2, %1
	jf	%4, bad
	data4	0x4E08000D	! bswap %2, %0, W-13/8
	shr	%3, %1, 13
	cmpe	%4, %2, %3
	jf	%4, bad

	lea	%0, [ctest]
	lea	%1, [c__end]
	sub	%1, 2*W
0:	ld	%2, [%0], 3
	adc	%5, %2, %3, %4		! 5A
	add	%6, %2, %3
	and	%7, %4, 1
	add	%6, %7
	cmpe	%7, %6, %5
	jf	%7, bad
	sbb	%5, %2, %3, %4		! 5B
	sub	%6, %2, %3
	and	%7, %4, 1
	sub	%6, %7
	cmpe	%7, %6, %5
	jf	%7, bad
	add	%0, W
	cmpb	%8, %0, %1
	jt	%8, 0b

	mov	%0, 0
	mov	%1, 1
	mov	%2, 2
	mov	%3, -2
	adcc	%4, %3, %0, %0		! 50
	cmpe	%4, %4, 0
	jf	%4, bad
	adcc	%4, %3, %0, %1
	cmpe	%4, %4, 0
	jf	%4, bad
	adcc	%4, %3, %1, %0
	cmpe	%4, %4, 0
	jf	%4, bad
	adcc	%4, %3, %1, %1
	cmpe	%4, %4, 1
	jf	%4, bad
	adcc	%4, %3, %2, %0
	cmpe	%4, %4, 1
	jf	%4, bad
	adcc	%4, %3, %2, %1
	cmpe	%4, %4, 1
	jf	%4, bad

	mov	%0, 0
	mov	%1, 1
	mov	%2, 2
	mov	%3, 1
	sbbb	%4, %3, %0, %0		! 51
	cmpe	%4, %4, 0
	jf	%4, bad
	sbbb	%4, %3, %0, %1
	cmpe	%4, %4, 0
	jf	%4, bad
	sbbb	%4, %3, %1, %0
	cmpe	%4, %4, 0
	jf	%4, bad
	sbbb	%4, %3, %1, %1
	cmpe	%4, %4, 1
	jf	%4, bad
	sbbb	%4, %3, %2, %0
	cmpe	%4, %4, 1
	jf	%4, bad
	sbbb	%4, %3, %2, %1
	cmpe	%4, %4, 1
	jf	%4, bad

	mov	%0, -12345
	mov	%1, -54321
	mulh	%2, %0, %1		! 52
	mov	%3, -66666
	cmpe	%4, %2, %3
	jf	%4, bad

	mov	%0, %flags
	or	%0, (1<<T_SUPER) - 1	! Enable all arithmetic traps
	mov	%flags, %0

	mov	%60, -1			! Reset both trap report registers
	mov	%61, -1
	divu	%0, %60, 0		! Division by zero
	cmpe	%1, %0, -1		! UNS_MAX?
	jf	%1, bad
	cmpe	%1, %60, -1		! User trap handler shouldn't care
	jf	%1, bad
	cmpe	%1, %61, T_IDIVZ	! Trap handler task should see it
	jf	%1, bad

	lea	%0, [ctrap_function]
	mov	%tfun, %0		! %tfun = trap_function
	mov	%0, %tmask
	or	%0, (1<<T_SUPER) - 1	! Trap all arithmetic traps locally
	mov	%tmask, %0

	mov	%61, -1
	divu	%0, %60, 0		! Division by zero
	cmpe	%1, %60, T_IDIVZ	! User trap handler should see it
	jf	%1, bad
	cmpe	%1, %61, -1		! Trap handler task no longer involved
	jf	%1, bad

	mov	%0, -4
	shr	%0, 1
	mov	%1, 1
	mov	%60, -1
	add	%2, %0, %1
	addo	%3, %0, %1		! 53
	jge	%60, bad
	cmpe	%4, %2, %3
	jf	%4, bad
	add	%2, %2, %1
	addo	%3, %3, %1
	cmpe	%4, %60, T_IOVFL
	jf	%4, bad
	cmpe	%4, %2, %3
	jf	%4, bad

	mov	%0, 1
	bis	%0, (W*8)-1
	mov	%1, 1
	mov	%60, -1
	sub	%2, %0, %1
	subo	%3, %0, %1		! 54
	jge	%60, bad
	cmpe	%4, %2, %3
	jf	%4, bad
	sub	%2, %2, %1
	subo	%3, %3, %1
	cmpe	%4, %60, T_IOVFL
	jf	%4, bad
	cmpe	%4, %2, %3
	jf	%4, bad

	mov	%0, 0
	bis	%0, (W*8)-3
	mov	%1, 2
	mov	%60, -1
	mul	%2, %0, %1
	mulo	%3, %0, %1		! 55
	jge	%60, bad
	cmpe	%4, %2, %3
	jf	%4, bad
	mul	%2, %2, %1
	mulo	%3, %3, %1
	cmpe	%4, %60, T_IOVFL
	jf	%4, bad
	cmpe	%4, %2, %3
	jf	%4, bad

	mov	%0, -5
	zxt	%1, %0, 10		! 56
	cmpe	%2, %1, -5 & 0x3FF
	jf	%2, bad
	sxt	%3, %1, 11		! 57
	cmpe	%4, %3, %1
	jf	%4, bad
	sxt	%4, %1, 10
	cmpe	%5, %4, %0
	jf	%5, bad
	mov	%60, -1
	sxto	%5, %0, 10		! 3D
	jge	%60, bad
	cmpe	%6, %5, %0
	jf	%6, bad
	sxto	%7, %1, 10
	cmpe	%8, %60, T_IOVFL
	jf	%8, bad
	cmpe	%8, %7, %0
	jf	%8, bad

	mov	%0, 123
	mov	%1, 234
	mov	%2, 100
1:	add4	%3, %0, %1		! 1B
	add	%4, %0, %1
	call	4f
	sub4	%3, %0, %1		! 1C
	sub	%4, %0, %1
	call	4f
	mul4	%3, %0, %1		! 1D
	mul	%4, %0, %1
	call	4f
	and	%5, %1, 31
	shl4	%3, %0, %5		! 20
	shl	%4, %0, %5
	call	4f
	and	%5, %1, 31
	sal4	%3, %0, %5		! 3E
	sal	%4, %0, %5
	sxt	%4, %4, 32
	call	5f
	and	%5, %1, 31
	sar4	%3, %0, %5		! 3F
	sxt	%4, %0, 32
	sar	%4, %4, %5
	call	5f
	add4	%3, %0, 1001		! 31
	add	%4, %0, 1001
	call	4f
	sub4	%3, %0, 1001
	sub	%4, %0, 1001
	call	4f
	mul4	%3, %0, 1001		! 23
	mul	%4, %0, 1001
	call	4f
	shl4	%3, %0, 16		! 24
	shl	%4, %0, 16
	call	4f
	sal4	%3, %0, 16		! 40
	sal	%4, %0, 16
	sxt	%4, %4, 32
	call	5f
	sar4	%3, %0, 16		! 43
	sxt	%4, %0, 32
	sar	%4, %4, 16
	call	5f
	sub4	%3, 1001, %0		! 27
	sub	%4, 1001, %0
	call	4f
	mul	%0, 3
	mul	%1, 5
	sub	%2, 1
	jnz	%2, 1b
	jmp	9f
4:					! See if %3 == zxt(%4, 32)
	zxt	%4, %4, 32
5:	cmpe	%5, %3, %4
	jf	%5, bad
	ret
9:
	lea	%30, [ctest]		! Start of this program
	lea	%31, [c__end]		! Start of heap
	mov	%32, 64			! Enough bytes to play with

	divu	%0, %32, 8*4		! Size 8, 4 times unrolled
	mov	%1, 2			! Index, displaced by 2
0:	ld8	%2, %30[%1-2]		! 60
	st8	%2, %31[%1-2]		! 61
	ld8	%2, %30[%1-1]
	st8	%2, %31[%1-1]
	ld8	%2, %30[%1+0]
	st8	%2, %31[%1+0]
	ld8	%2, %30[%1+1]
	st8	%2, %31[%1+1]
	add	%1, 4
	sub	%0, 1
	jnz	%0, 0b
	call	checkldst
	jnz	%0, bad

	divu	%0, %32, 4*4		! Size 4, 4 times unrolled
	mov	%1, 2			! Index, displaced by 2
0:	ld4u	%2, %30[%1-2]		! 62
	st4	%2, %31[%1-2]		! 64
	ld4s	%2, %30[%1-1]		! 63
	st4	%2, %31[%1-1]
	ld4u	%2, %30[%1+0]
	st4	%2, %31[%1+0]
	ld4s	%2, %30[%1+1]
	st4	%2, %31[%1+1]
	add	%1, 4
	sub	%0, 1
	jnz	%0, 0b
	call	checkldst
	jnz	%0, bad

	divu	%0, %32, 2*4		! Size 2, 4 times unrolled
	mov	%1, 2			! Index, displaced by 2
0:	ld2u	%2, %30[%1-2]		! 65
	st2	%2, %31[%1-2]		! 67
	ld2s	%2, %30[%1-1]		! 66
	st2	%2, %31[%1-1]
	ld2u	%2, %30[%1+0]
	st2	%2, %31[%1+0]
	ld2s	%2, %30[%1+1]
	st2	%2, %31[%1+1]
	add	%1, 4
	sub	%0, 1
	jnz	%0, 0b
	call	checkldst
	jnz	%0, bad

	divu	%0, %32, 1*4		! Size 1, 4 times unrolled
	mov	%1, 2			! Index, displaced by 2
0:	ld1u	%2, %30[%1-2]		! 68
	st1	%2, %31[%1-2]		! 6A
	ld1s	%2, %30[%1-1]		! 69
	st1	%2, %31[%1-1]
	ld1u	%2, %30[%1+0]
	st1	%2, %31[%1+0]
	ld1s	%2, %30[%1+1]
	st1	%2, %31[%1+1]
	add	%1, 4
	sub	%0, 1
	jnz	%0, 0b
	call	checkldst
	jnz	%0, bad

	add	%40, %30, (1<<10)	! Our two areas, displaced.
	add	%41, %31, -(1<<10)

	mov	%32, 8
	ld8	%0, [%40-(1<<10)]	! 6C
	st8	%0, [%41+(1<<10)]	! 6D
	call	checkldst
	jnz	%0, bad

	mov	%32, 2*4
	ld4u	%0, [%40-(1<<10)]	! 6E
	st4	%0, [%41+(1<<10)]	! 70
	ld4s	%0, [%40-(1<<10)+4]	! 6F
	st4	%0, [%41+(1<<10)+4]
	call	checkldst
	jnz	%0, bad

	mov	%32, 2*2
	ld2u	%0, [%40-(1<<10)]	! 71
	st2	%0, [%41+(1<<10)]	! 73
	ld2s	%0, [%40-(1<<10)+2]	! 72
	st2	%0, [%41+(1<<10)+2]
	call	checkldst
	jnz	%0, bad

	mov	%32, 2*1
	ld1u	%0, [%40-(1<<10)]	! 74
	st1	%0, [%41+(1<<10)]	! 76
	ld1s	%0, [%40-(1<<10)+1]	! 75
	st1	%0, [%41+(1<<10)+1]
	call	checkldst
	jnz	%0, bad

	mov	%0, 0
	lea	%1, [cc0123]
	ld	%2, [%1]
	ld	%3, [%1+W]
	mov	%4, 0
0:	add	%1, 1
	add	%4, 8
	ld	%5, [%1]		! unaligned word load
	ld	%6, %1[%0]		! ditto
	cmpe	%6, %6, %5
	jf	%6, bad
	shr	%6, %2, %4
	sub	%7, W*8, %4
	shl	%7, %3, %7
	or	%6, %7
	cmpe	%5, %5, %6
	jf	%5, bad
	cmpb	%5, %4, (W-1)*8
	jt	%5, 0b

	mov	%10, 0xBEEF
	mov	%11, 0xDEAD
	shl	%12, %11, 16
	or	%12, %10		! %12 = 0xDEADBEEF
	shl	%13, %10, 16
	or	%13, %11		! %13 = 0xBEEFDEAD

	ld8	%0, [cDEAD]		! 78
	mov	%10, W-4		! W = 4?
	jnz	%10, 8f
4:	cmpe	%2, %0, %12		! %0:%1 = 0xBEEFDEADDEADBEEF?
	jf	%2, bad
	cmpe	%2, %1, %13
	jf	%2, bad
	jmp	9f
8:	shl	%2, %13, 32
	or	%2, %12			! %2 = 0xBEEFDEADDEADBEEF
	cmpe	%2, %0, %2		! %0 = 0xBEEFDEADDEADBEEF?
	jf	%2, bad
9:
	ld4u	%0, [cDEAD]		! 79
	cmpe	%1, %0, %12		! %0 = 0xDEADBEEF?
	jf	%1, bad

	ld4s	%0, [cDEAD]		! 7A
	mov	%2, %12
	jz	%10, 4f
	sxt	%2, %2, 32
4:	cmpe	%1, %0, %2		! %0 = sxt(0xDEADBEEF, 32)?
	jf	%1, bad

	!lea tested elsewhere		! 77

	mov	%32, 30*W
	ld	%0, [%30], 30		! 7B
	st	%0, [%31], 30		! 7C
	call	checkldst
	jnz	%0, bad

	lea	%32, [c__end]
	lea	%0, [ctest]
	sub	%32, %0			! Size of entire program
	mov	%0, %31
	mov	%1, %30
	mov	%2, %32
	copy	%0, %1, %2		! 7D
	call	checkldst
	jnz	%0, bad

	!fill tested elsewhere		! 7E

	mov	%0, %31
	mov	%1, 'K'			! Something we won't find
	mov	%2, %32
	scan	%0, %1, %2		! 7F
	jnz	%2, bad			! Count should be zero
	add	%3, %31, %32
	cmpe	%3, %3, %0		! Pointer at end
	jf	%3, bad
	mov	%0, %31
	mov	%1, 'K'
	mov	%2, %32
	st1	%1, [%0+123]		! Now we will find something
	scan	%0, %1, %2
	sub	%3, %32, 123
	cmpe	%3, %3, %2		! Leftover count
	jf	%3, bad
	add	%3, %31, 123
	cmpe	%3, %3, %0		! Pointer at character found
	jf	%3, bad

	!in*, out*, can't test		! 80 - 87

	mov	%0, 123
	mov	%64, %0			! 89
	mov	%1, %64			! 88
	cmpe	%2, %0, %1
	jf	%2, bad

	mov	%1, %sp
	sub	%sp, 123*W
	mov	%0, 0
	call	9f
	jmp	0f
c_deep:	data4	(123 + 1 + (FP-1+1) + (1<<17)) * W
9:
	enter	%1, (1<<17)*W		! 8A
	mov	%1, %0
	mov	%0, %sp
	jnz	%1, 8f
	return	%1, 123*W		! 8C
8:	leave	%1			! 8B
	ret	321*W			! 8D

0:	cmpe	%2, %1, %sp
	jf	%2, bad
	ld4u	%2, [c_deep]
	sub	%2, %sp, %2
	cmpe	%2, %0, %2
	jf	%2, bad
	sub	%sp, 321*W
	mov	%0, 1
	call	9b
	cmpe	%2, %1, %sp
	jf	%2, bad

	!queue tested elsewhere		! 8E

	!trap tested elsewhere		! 8F

	fmov	%10, 1.5
	fmov	%12, 2.25
	fmov	%14, -3.75

	fadd	%0, %10, %12		! 90
	fcmpe	%2, %0, 3.75
	jf	%2, bad
	fsub	%0, %10, %12		! 91
	fcmpe	%2, %0, -0.75
	jf	%2, bad
	fmul	%0, %10, %12		! 92
	fmov	%2, 3.375
	fcmpe	%4, %0, %2
	jf	%4, bad
	fdiv	%0, %14, %10		! 93
	fcmpe	%2, %0, -2.5
	jf	%2, bad
	frem	%0, %14, %10		! 94
	fcmpe	%2, %0, -0.75
	jf	%2, bad
	fcmpg	%0, %10, %12		! 96
	jt	%0, bad
	fcmpl	%0, %10, %12
	jf	%0, bad
	fcmpg	%0, %10, %10
	jt	%0, bad
	fcmpge	%0, %10, %12		! 97
	jt	%0, bad
	fcmple	%0, %10, %12
	jf	%0, bad
	fcmpge	%0, %10, %10
	jf	%0, bad
	fcmpe	%0, %10, %10		! 99
	jf	%0, bad
	fcmpne	%0, %10, %10		! 9A
	jt	%0, bad
	fcmpe	%0, %10, %12
	jt	%0, bad
	fcmpne	%0, %10, %12
	jf	%0, bad
	mov	%0, 5
	finsx	%2, %10, %0		! 9A
	fcmpe	%4, %2, 48
	jf	%4, bad

	fadd	%0, %10, 2.25		! A0
	fcmpe	%2, %0, 3.75
	jf	%2, bad
	fsub	%0, %10, 2.25		! A1
	fcmpe	%2, %0, -0.75
	jf	%2, bad
	fmul	%0, %10, 2.25		! A2
	fmov	%2, 3.375
	fcmpe	%4, %0, %2
	jf	%4, bad
	fdiv	%0, %14, 1.5		! A3
	fcmpe	%2, %0, -2.5
	jf	%2, bad
	frem	%0, %14, 1.5		! A4
	fcmpe	%2, %0, -0.75
	jf	%2, bad
	fcmpg	%0, %10, 2.25		! A6
	jt	%0, bad
	fcmpl	%0, %10, 2.25		! A5
	jf	%0, bad
	fcmpg	%0, %10, 1.5		! A6
	jt	%0, bad
	fcmpge	%0, %10, 2.25		! A8
	jt	%0, bad
	fcmple	%0, %10, 2.25		! A7
	jf	%0, bad
	fcmpge	%0, %10, 1.5		! A8
	jf	%0, bad
	fcmpe	%0, %10, 1.5		! A9
	jf	%0, bad
	fcmpne	%0, %10, 1.5		! AA
	jt	%0, bad
	fcmpe	%0, %10, 2.25
	jt	%0, bad
	fcmpne	%0, %10, 2.25
	jf	%0, bad
	finsx	%2, %10, 5		! AB
	fcmpe	%4, %2, 48
	jf	%4, bad
	fsub	%0, 2.25, %10		! AC
	fcmpe	%2, %0, 0.75
	jf	%2, bad
	fdiv	%0, -3.75, %10		! A3
	fcmpe	%2, %0, -2.5
	jf	%2, bad

	! fmov tested elsewhere		! B0

	fextx	%0, %14			! B1
	cmpe	%1, %0, 2
	jf	%1, bad
	fround	%0, %10			! B2
	fcmpe	%2, %0, 2
	jf	%2, bad
	fneg	%0, %10
	fround	%2, %0
	fcmpe	%4, %2, -2
	jf	%4, bad
	fmov	%0, 2.5
	fround	%2, %0
	fcmpe	%4, %2, 2
	jf	%4, bad
	ftrunc	%0, %10			! B3
	fcmpe	%2, %0, 1
	jf	%2, bad
	ftrunc	%0, %14
	fcmpe	%2, %0, -3
	jf	%2, bad

	! ill tested when tests failed	! 44

	! call tested elsewhere		! Fx

done:	trap	%0, %0, T_U_EXIT	! All OK?  Exit!

checkldst:				! Compare %30[] with %31[] length %32
	enter	%59, 0			! %0 == 0 if OK
	mov	%0, %32			! Wipe %31[] when done
	mov	%61, 0
0:	ld1u	%60, %30[%61]
	ld1u	%59, %31[%61]
	cmpe	%60, %60, %59
	jf	%60, 1f
	add	%61, 1
	sub	%0, 1
	jnz	%0, 0b
	mov	%61, %31
	mov	%60, 0
	mov	%59, %32
	fill	%61, %60, %59
1:	return	%59

	align	W
ctest:	dataw	test-.
ctrap_function:dataw	trap_function-.
c__etext:dataw	__end-.
c__end:	dataw	__end-.
cc0123:	dataw	c0123-.
cDEAD:	data4	0xDEADBEEF
	data4	0xBEEFDEAD
c0123:	data4	0x03020100
	data4	0x07060504
	data4	0x0B0A0908
	data4	0x0F0E0D0C

	align	W
__etext:
__edata:
__end:
