!	cpqboot 1.1 - Prebootstrap for picky systems like a particular Compaq
!							Author: Kees J. Bot
!								17 Feb 2001
! This bootstrap looks as much as possible like a plain old
! MS-DOS bootstrap, except that its only function is to wait for you
! to insert another boot diskette and press ENTER.  This is needed to
! circumvent systems whose BIOS performs a check to see if what it is
! booting is proper Redmond based operating system as God intended.

	LOADOFF	   =	0x7C00	! 0x0000:LOADOFF is where this code is loaded
	BUFFER	   =	0x0600	! First free memory

.sect .text; .sect .rom; .sect .data; .sect .bss
.sect .text

! Start boot procedure.

boot:
	jmp	over		! Jump over the "make 'em happy" junk
	nop

.ascii	"CPQBOOT0"		! OEM name and version
.data2	   2			! Bytes per sector
.data1	   1			! Sectors per cluster
.data2	   1			! Reserved (boot) sectors
.data1	   2			! Number of FAT tables
.data2	 224			! Number of root directory entries
.data2	2880			! Number of sectors if < 32M
.data1	0xF0			! Media descriptor
.data2	   9			! Sectors per FAT if FAT-12 or FAT-16
.data2	  18			! Sectors per track
.data2	   2			! Number of heads
.data4	   0			! Hidden sectors
.data4	   0			! Number of sectors if >= 32M
.data1	   0			! Physical drive?
.data1	   0			! Reserved
.data1	  41			! DOS > 4.0 diskette
.ascii	"KJB "			! Serial number
.ascii	"was here!  "		! Volume label
.ascii	"FAT12   "		! FAT type

over:
	xor	ax, ax		! ax = 0x0000, the vector segment
	mov	ds, ax
	mov	es, ax
	cli			! Ignore interrupts while setting stack
	mov	ss, ax		! ss = es = ds = vector segment
	mov	sp, LOADOFF	! Usual place for a bootstrap stack
	sti

	mov	si, LOADOFF	! si = start of this code
	mov	di, BUFFER	! Buffer area
	mov	cx, 512/2	! One sector
	cld
    rep	movs
	jmpf	0:BUFFER+migrate ! To safety
migrate:

	testb	dl, dl		! Winchester disks if dl >= 0x80
	jns	floppy

winchester:
	mov	si, BUFFER+whyhd
	call	print
	jmp	hang
.sect .data
whyhd:	.ascii	"This code should be booted from a floppy. \0"
.sect .text

floppy:
	mov	si, BUFFER+nextboot
	call	print

	xorb	ah, ah		! Read character from keyboard
	int	0x16

.sect .data
nextboot:
.ascii	"Please insert a proper boot diskette and hit a key to boot it.\r\n\0"
.sect .text

	mov	di, 3		! Retry count
load:
	mov	bx, LOADOFF	! Load first sector into the usual place
	mov	ax, 0x0201	! Read sector, #sectors = 1
	mov	cx, 0x0001	! Track 0, sector 1
	xorb	dh, dh		! Drive dl, head 0
	int	0x13
	jnc	done		! Read OK?
	dec	di
	jz	error		! Too many retries?

	xorb	ah, ah		! Reset drive
	int	0x13
	jmp	load		! Try again

done:

! Call new bootstrap.
	jmpf	0:LOADOFF

! Read error: print message, hang forever
error:
	mov	si, BUFFER+errno+1
prnum:	movb	al, ah		! Error number in ah
	andb	al, 0x0F	! Low 4 bits
	cmpb	al, 10		! A-F?
	jb	digit		! 0-9!
	addb	al, 7		! 'A' - ':'
digit:	addb	(si), al	! Modify '0' in string
	dec	si
	movb	cl, 4		! Next 4 bits
	shrb	ah, cl
	jnz	prnum		! Again if digit > 0

	mov	si, BUFFER+rderr  ! String to print
	call	print

! Hang forever waiting for CTRL-ALT-DEL
hang:	hlt
	jmp	hang

0:	movb	ah, 0x0E	! Print character in teletype mode
	mov	bx, 0x0001	! Page 0, foreground color
	int	0x10		! Call BIOS VIDEO_IO
print:	lodsb			! al = *si++ is char to be printed
	testb	al, al
	jnz	0b
	ret


.sect .data
rderr:	.ascii	"Read error "
errno:	.ascii	"00 \0"
