!	xtboot 1.1 - XT floppy parameters installer.
!							Author: Kees J. Bot
!								18 Mar 1999

	LOADOFF	   =	0x7C00	! 0x0000:LOADOFF is where this code is loaded
	BUFFER	   =	0x0600	! First free memory
	DSKBASE    =	  0x1E	! Floppy disk parameter vector
	DSKPARSIZE =	    11	! 11 bytes of floppy parameters
	DSKPARBUF  =    0x05F0	! Out of the way place for the parameters
	SECTORS	   =	     4	! Offset into parameters to sectors per track

.sect .text; .sect .rom; .sect .data; .sect .bss
.sect .text

! Start boot procedure.

boot:
	xor	ax, ax		! ax = 0x0000, the vector segment
	mov	ds, ax
	mov	es, ax
	cli			! Ignore interrupts while setting stack
	mov	ss, ax		! ss = es = ds = vector segment
	mov	sp, LOADOFF	! Usual place for a bootstrap stack
	sti

	mov	si, LOADOFF	! si = start of this code
	mov	di, BUFFER	! Buffer area
	mov	cx, 512/2	! One sector
	cld
    rep	movs
	jmpf	0:BUFFER+migrate ! To safety
migrate:

	mov	si, BUFFER+parameters	! char (*si)[DSKPARSIZE] = parameters;

	testb	dl, dl		! Winchester disks if dl >= 0x80
	jge	floppy

winchester:

	mov	si, BUFFER+whyhd
	call	print
	jmp	hang
.sect .data
whyhd:	.ascii	"This code should be booted from a floppy. \0"
.sect .text

! Floppy:
! Execute three read tests to determine the drive type.  Test for each floppy
! type by reading the last sector on the first track.  If it fails, try a type
! that has less sectors.  Therefore we start with 1.44M (18 sectors) then 1.2M
! (15 sectors) ending with 720K/360K (both 9 sectors).  (The floppy parameters
! of the last two are equal, apart from the motor start time.  This saves us
! the rather painful "try to read track 41" test.)

floppy:
	mov	di, DSKPARBUF	! Put parameters in "reserved for DOS"
	mov	(DSKBASE*4+0), di	! Load offset of disk parameters
	mov	(DSKBASE*4+2), ds	! Load segment of disk parameters
	mov	cx, DSKPARSIZE
    rep	movsb			! Copy next set of parameters in place

	xorb	ah, ah		! Reset drive
	int	0x13

	movb	cl, (DSKPARBUF+SECTORS)	! cl = number of last sector on track

	cmpb	cl, 9 		 ! No need to do the last 720K/360K test
	je	success

! Try to read the last sector on track 0

	mov	bx, LOADOFF	! es:bx = scratch buffer
	mov	ax, 0x0201	! Read sector, #sectors = 1
	xorb	ch, ch		! Track 0, last sector
	xorb	dh, dh		! Drive dl, head 0
	int	0x13
	jb	floppy		! Error, try the next floppy type
success:

	mov	si, BUFFER+nextboot
	call	print

	xorb	ah, ah		! Read character from keyboard
	int	0x16

.sect .data
nextboot:
.ascii	"Floppy parameters have been installed.  Please insert a ROOT\r\n"
.ascii	"diskette of the same type and hit a key to retry booting.\r\n\0"
.sect .text

	mov	di, 3		! Retry count
load:
	mov	bx, LOADOFF	! Load first sector into the usual place
	mov	ax, 0x0201	! Read sector, #sectors = 1
	mov	cx, 0x0001	! Track 0, sector 1
	xorb	dh, dh		! Drive dl, head 0
	int	0x13
	jnc	done		! Read OK?
	dec	di
	jz	error		! Too many retries?

	xorb	ah, ah		! Reset drive
	int	0x13
	jmp	load		! Try again

done:

! Call new bootstrap.
	jmpf	0:LOADOFF

! Read error: print message, hang forever
error:
	mov	si, BUFFER+errno+1
prnum:	movb	al, ah		! Error number in ah
	andb	al, 0x0F	! Low 4 bits
	cmpb	al, 10		! A-F?
	jb	digit		! 0-9!
	addb	al, 7		! 'A' - ':'
digit:	addb	(si), al	! Modify '0' in string
	dec	si
	movb	cl, 4		! Next 4 bits
	shrb	ah, cl
	jnz	prnum		! Again if digit > 0

	mov	si, BUFFER+rderr  ! String to print
	call	print

! Hang forever waiting for CTRL-ALT-DEL
hang:	jmp	hang

0:	movb	ah, 0x0E	! Print character in teletype mode
	mov	bx, 0x0001	! Page 0, foreground color
	int	0x10		! Call BIOS VIDEO_IO
print:	lodsb			! al = *si++ is char to be printed
	testb	al, al
	jnz	0b
	ret


.sect .data
rderr:	.ascii	"Read error "
errno:	.ascii	"00 \0"

parameters:
! Floppy disk parameters sorted down by sectors per track.  (The format gap
! length params of the 3.5" disks might be wrong, but that won't matter.)

! 1.44M 3.5"
dshd3:	.data1	0xAF, 0x02, 25, 2, 18, 0x1B, 0xFF, 0x54, 0xF6, 15, 8

! 1.2M 5.25"
dshd5:	.data1	0xDF, 0x02, 25, 2, 15, 0x1B, 0xFF, 0x54, 0xF6, 15, 8

! 720K 3.5", also used for 360K 5.25"
dsdd3:	.data1	0xDF, 0x02, 25, 2,  9, 0x2A, 0xFF, 0x50, 0xF6, 15, 8

! Just for completeness, here are the real 360K params.
!dsdd5:	.data1	0xDF, 0x02, 25, 2,  9, 0x2A, 0xFF, 0x54, 0xF6, 15, 3
